package com.japisoft.dtdparser.node;

import java.io.*;
import java.util.Hashtable;
import java.util.Enumeration;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class ElementDTDNode extends DTDNode implements DocumentWritable {

	public ElementDTDNode() {
		super();
		setNodeType( ELEMENT );
	}

	/** Build an Element node with the following name */
	public ElementDTDNode( String name ) {
		this();
		setName( name );
	}

	private Hashtable htAttribute;
	private String name;

	private String attComment = null;
	
	public void addAttComment( String attComment ) {
		if ( this.attComment == null )
			this.attComment = attComment;
		else
			this.attComment += "\n" + attComment;
	}

	/** Overriding for getting attribute node */
	public void addDTDNode(DTDNode node) {
		if (node.isAttribute()) {
			if (htAttribute == null) {
				htAttribute = new Hashtable();
			}
			AttributeDTDNode n = ( AttributeDTDNode ) node;
			if ( attComment != null )
				n.addNodeComment( attComment );
			
			if  ( n.getName() == null )  {
			} else
			htAttribute.put(n.getName(), n);
		}
		super.addDTDNode(node);
	}

	/** @return a DTD attribute declaration for the following name */
	public AttributeDTDNode getAttributeDeclaration(String name) {
		if (htAttribute == null)
			return null;
		return (AttributeDTDNode) htAttribute.get(name);
	}

	/** Name of the element */
	public void setName( String name ) {
		this.name = name;
	}

	/** @return the name of the element ( tag name ) */
	public String getName() {
		return name;
	}

	private boolean empty = false;

	/** empty node : by default false */
	public void setEmptyElement( boolean empty ) {
		this.empty = empty;
	}

	/** @return true if the node is empty */
	public boolean isEmptyElement() {
		return empty;
	}

	/** default operator */
	public final static int ONE_ITEM_OPERATOR = 0;
	/** '?' operator */
	public final static int ZERO_ONE_ITEM_OPERATOR = 1;
	/** '*' operator */
	public final static int ZERO_MORE_ITEM_OPERATOR = 2;
	/** '+' operator */
	public final static int ONE_MORE_ITEM_OPERATOR = 3;

	private int op = ONE_ITEM_OPERATOR;

	/** Set the occurency operator :
	- ZERO_MORE_ITEM_OPERATOR,
	- ONE_MORE_ITEM_OPERATOR,
	- ZERO_ONE_ITEM_OPERATOR */
	public void setOperator(int op) {
		this.op = op;
	}

	/** @return "*", "+", "?" or "" */
	public String getOperatorString() {
		switch ( op ) {
			case ZERO_MORE_ITEM_OPERATOR :
				return "*";
			case ONE_MORE_ITEM_OPERATOR :
				return "+";
			case ZERO_ONE_ITEM_OPERATOR :
				return "?";
		}
		return "";
	}

	/** @return the occurency operator */
	public int getOperator() {
		return op;
	}

	private boolean pcdata = false;

	/** PCDATA support for text content */
	public void setPCDATA(boolean pcdata) {
		this.pcdata = pcdata;
	}

	/** PCDATA support for text content */
	public boolean hasPCDATA() {
		return pcdata;
	}

	private boolean any = false;

	/** Can have any element children */
	public void setANY(boolean any) {
		this.any = any;
	}

	/** Can have any element children */
	public boolean hasANY() {
		return any;
	}

	/** @return <code>true</code> if the nodeName is supported as child, 
	previousName is to check if the sequence if a good one,
	occurrence if for the child count */
	public boolean isNodeChildSupported(
		String nodeName,
		String previousName,
		int occurrence) {

		if (isEmptyElement())
			return false;
		if (hasANY())
			return true;
		if (hasPCDATA())
			return false;
		for (int i = 0; i < getDTDNodeCount(); i++) {
			if (getDTDNodeAt(i) instanceof ElementSetDTDNode) {
				ElementSetDTDNode set = (ElementSetDTDNode) getDTDNodeAt(i);

				return set.isNodeChildSupported(
					nodeName,
					previousName,
					occurrence);
			}
		}

		return false;
	}

	/** Write a minimal valid XML document */
	public void writeDocument(PrintWriter output) throws IOException {
		
		output.write("<" + getName());

		// Write attribute
		Enumeration enume = getDTDNodeForType(DTDNode.ATTRIBUTE);
		while (enume.hasMoreElements()) {
			AttributeDTDNode att = (AttributeDTDNode) enume.nextElement();
			if (att.getUsage() == AttributeDTDNode.REQUIRED_ATT) {
				output.write(" ");
				output.write(att.getName());
				output.write("=\"");
				output.write(att.getDefaultValue());
				output.write("\"");
			}
		}

		if (isEmptyElement()) {
			output.write("/>\n");
		} else {
			output.write(">\n");
			enume = getDTDNodeForType(DTDNode.ELEMENT_SET);
			if (enume.hasMoreElements()) {
				ElementSetDTDNode set = (ElementSetDTDNode) enume.nextElement();
				set.writeDocument(output);
			}
			output.write("</" + getName() + ">\n"); // ANY ?
		}
	}

	public String toString() {
		StringBuffer sb = new StringBuffer();
		
		sb.append( getDTDComment() );
		
		
		sb.append("<!ELEMENT ").append(getName()).append(" ");
		if (isEmptyElement())
			sb.append("EMPTY>");
		else if (hasANY())
			sb.append("ANY>");
		else if (hasPCDATA())
			sb.append("(#PCDATA)>");
		else {
			// Show the set once
			Enumeration enume = getDTDNodeForType(DTDNode.ELEMENT_SET);
			if (enume.hasMoreElements())
				sb.append("" + enume.nextElement());
			sb.append(">");
		}

		// Show the attributes
		Enumeration enume = getDTDNodeForType(DTDNode.ATTRIBUTE);
		if (enume.hasMoreElements()) {
			sb.append("\n");
			sb.append( getDTDComment( attComment ) );
			sb.append("<!ATTLIST " + getName());
			sb.append(" \n");
			while (enume.hasMoreElements()) {
				AttributeDTDNode node = (AttributeDTDNode) enume.nextElement();
				sb.append(node.toString());
				sb.append("\n");
			}

			sb.append(">");
		}

		return sb.toString();
	}

}

// ElementDTDNode ends here
