package com.japisoft.editix.document;

import java.io.File;

import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Icon;

import com.japisoft.editix.main.EditixApplicationModel;
import com.japisoft.editix.ui.EditixFactory;
import com.japisoft.framework.xml.XMLToolkit;
import com.japisoft.framework.xml.parser.FPParser;
import com.japisoft.framework.xml.parser.node.FPNode;
import com.japisoft.xmlpad.XMLDocumentInfo;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class TemplateModel {

	private static final String TEMPLATE_USER = "template/user/";
	private static final String DEFAULT_ENCODING = "UTF-8";
	private static List<GroupTemplate> model = null;

	static {
		loadModel();
	}

	public static void loadModel() {
		model = null;
		InputStream input = 
			ClassLoader.getSystemResourceAsStream( "templates.xml" );
		if (input == null) {
			System.err.println( "Can't find templates.xml" );
			System.exit( 1 );
		}
		try {
			loadModel( new InputStreamReader( input, DEFAULT_ENCODING ) );
			// Search from the user path
			File file = getCustomTemplatesPath();
			if ( file.exists() ) {
				try {
					loadModel(
						new InputStreamReader(
							new FileInputStream( file ), DEFAULT_ENCODING )
					);
				} catch( Throwable th ) {
					th.printStackTrace();
				}
			}
		} catch( Exception exc ) {
			System.err.println( "Can't load templates.xml : " + exc.getMessage() );
		}
	}

	private static File getCustomTemplatesPath() {
		File userPath = EditixApplicationModel.getAppUserPath();
		File file = new File( userPath, "templates.xml" );
		return file;
	}

	static GroupTemplate getUserGroupTemplate() {
		for ( int i = 0; i < getGroupTemplateCount(); i++ ) {
			if ( "User".equalsIgnoreCase( getGroupTemplate( i ).getName() ) )
				return getGroupTemplate( i );
		}
		return new GroupTemplate( 
				"User", 
				"images/text.png", 
				"images/document_plain.png" 
		);
	}
	
	public static TemplateInfo getTemplateByLabel( String label ) {
		for ( int i = 0; i < getGroupTemplateCount(); i++ ) {
			GroupTemplate gt = getGroupTemplate( i );
			for ( int j = 0; j < gt.getTemplateInfoCount(); j++ ) {
				TemplateInfo ti = gt.getTemplateInfo( j );
				if ( label.equalsIgnoreCase( ti.label ) )
					return ti;
			}
		}
		return null;
	}

	public static TemplateInfo getTemplateByType( String type ) {
		for ( int i = 0; i < getGroupTemplateCount(); i++ ) {
			GroupTemplate gt = getGroupTemplate( i );
			for ( int j = 0; j < gt.getTemplateInfoCount(); j++ ) {
				TemplateInfo ti = gt.getTemplateInfo( j );
				if ( type.equalsIgnoreCase( ti.type ) )
					return ti;
			}
		}
		return null;
	}
	
	public static void loadModel( Reader input ) {

		FPParser p = new FPParser();

		try {
			FPNode sn = ( FPNode )p.parse(input).getRoot();

			if ( model == null )
				model = new ArrayList<GroupTemplate>();

			for ( int i = 0; i < sn.childCount(); i++ ) {
				
				FPNode c = sn.childAt( i );
				
				if ( c.matchContent( "group" ) ) {
					
					GroupTemplate gt = buildGroupTemplate( c ); 
					
					model.add( gt );
				}
			}

			GroupTemplate user = getUserGroupTemplate();

			// Old user template
			
			for ( int i = 0; i < sn.childCount(); i++ ) {
				if ( sn.childAt( i ).matchContent( "template" ) ) {
					TemplateInfo ti = buildTemplateInfo( sn.childAt( i ) );
					ti.icon = getGroupIconByType( ti.type );
					user.addTemplate( ti );
				}
			}
			
			// New user template
			
			File users = getUserTemplates();
			String[] content = users.list();
			if ( content != null ) {
				for ( String c : content ) {
					int i = c.lastIndexOf( "." );
					if ( i > -1 ) {
						String templateName = c.substring( 0, i );
						String type = c.substring( i + 1 );
						TemplateInfo ti = new TemplateInfo();
						ti.label = templateName;
						ti.type = type;
						ti.location = TEMPLATE_USER + c;
						ti.icon = getGroupIconByType( type );
						user.addTemplate( ti );
					}
				}
			}

		} catch ( Exception exc ) {
			System.err.println( "Can't parse templates.xml : "
					+ exc.getMessage() + " !");
		}
	}

	static GroupTemplate buildGroupTemplate( FPNode node ) {
		GroupTemplate gt = new GroupTemplate( 
			node.getAttribute( "label" ), 
			node.getAttribute( "icon" ), 
			node.getAttribute( "docIcon" ) 
		);
		for ( int i = 0; i < node.childCount(); i++ ) {
			FPNode sn = node.childAt( i );
			if ( sn.matchContent( "template" ) )
				gt.addTemplate( buildTemplateInfo( sn ) );
		}
		return gt;
	}
	
	static Icon getGroupIconByType( String type ) {
		for ( int i = 0; i < getGroupTemplateCount(); i++ ) {
			GroupTemplate gt = getGroupTemplate( i );
			for ( int j = 0; j < gt.getTemplateInfoCount(); j++ ) {
				if ( type.equalsIgnoreCase( gt.getTemplateInfo( j ).type ) ) {
					return gt.getDocIcon();
				}
			}
		}
		return null;
	}

	static Icon WIZARD_ICON = null;
	
	static TemplateInfo buildTemplateInfo( FPNode node ) {
		TemplateInfo ti = new TemplateInfo();
		ti.label = node.getAttribute( "label" );
		ti.location = node.getAttribute( "location" );
		ti.type = node.getAttribute( "type" );
		ti.system = "true".equals( node.getAttribute( "system" ) );
		ti.encoding = node.getAttribute( "encoding" );
		ti.defDTDLocation = node.getAttribute( "defDTDLocation" );
		ti.defDTDRoot = node.getAttribute( "defDTDRoot" );
		ti.help = node.getAttribute( "help" );
		ti.wizard = node.getAttribute( "wizard" );
		
		XMLDocumentInfo info = DocumentModel.getDocumentForType( ti.type );
		if ( info != null )
			ti.icon = info.getDocumentIcon();
		
		if ( ti.wizard != null ) {
			if ( WIZARD_ICON == null )
			WIZARD_ICON = EditixFactory.getImageIcon( "images/astrologer.png" );
			if ( WIZARD_ICON != null )
				ti.icon = WIZARD_ICON;
		}
		
		return ti;
	}
	
	public static File getTemplatePath( String location ) {
		File userPath = EditixApplicationModel.getAppUserPath();
		File f = new File( userPath, location );
		if ( !f.getParentFile().exists() )
			f.getParentFile().mkdirs();
		return f;
	}
	
	public static File getUserTemplates() {
		File users = EditixApplicationModel.getAppUserPath();
		File templates = new File( users, TEMPLATE_USER );
		if ( !templates.exists() )
			templates.mkdirs();
		return templates;
	}

	public static File createNewTemplate( String name, String type, String content ) throws Exception {
		File templates = getUserTemplates();
		name = name.replace( " ", "_" );
		File template = new File( templates, name + "." + type );
		XMLToolkit.save( template, content );
		TemplateInfo ti = new TemplateInfo();
		ti.label = name;
		ti.type = type;
		ti.location = TEMPLATE_USER + name + "." + type;
		ti.icon = getGroupIconByType( type );
		getUserGroupTemplate().addTemplate( ti );
		return template;
	}

	public static void resolveTemplate( 
			String encoding, 
			String location, 
			XMLDocumentInfo info ) throws Throwable {

		File userPath = EditixApplicationModel.getAppUserPath();
		File f = new File( userPath, location );

		InputStream input = null;
		if ( f.exists() ) {
			input = new FileInputStream( f );
		} else {
			f = new File( location );
			if ( f.exists() ) {
				input = new FileInputStream( f );
			} else {			
				input = ClassLoader.getSystemResourceAsStream( location );
				if ( input == null ) {
					URL url = new URL( location );
					input = url.openStream();
				}
			}
		}
		info.setTemplate( 
				com.japisoft.framework.xml.XMLToolkit.getContentFromInputStream( 
						input, encoding ).getContent() );
	}
	
	public static int getGroupTemplateCount() {
		if ( model == null )
			return 0;
		return model.size();
	}

	public static GroupTemplate getGroupTemplate( int index ) {
		if ( model == null )
			return null;
		return model.get( index );
	}

}
