package com.japisoft.editix.plugin;

import java.io.StringReader;
import java.io.StringWriter;

import javax.swing.text.JTextComponent;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;

import com.japisoft.editix.action.file.SaveAction;
import com.japisoft.editix.ui.EditixFrame;
import com.japisoft.xmlpad.XMLContainer;
/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class EditixDocument {

	private int index;

	EditixDocument( int index ) {
		this.index = index;
	}

	/**
	 * @return The document type (DTD, CSS, XML, XSD...) */
	public String getType() {
		XMLContainer c = getContainer();
		if ( c == null )
			return null;
		return c.getDocumentInfo().getType();
	}

	/**
	 * @return The document path */
	public String getLocation() {
		XMLContainer c = getContainer();
		if ( c == null )
			return null;
		return c.getCurrentDocumentLocation();
	}
	
	/** @return the caret position from 0 */
	public int getCaretLocation() {
		XMLContainer c = getContainer();
		if ( c == null )
			return 0;
		return c.getEditor().getCaretPosition();		
	}

	/**
	 * @param location Set a default location
	 */
	public void setLocation( String location ) {
		XMLContainer c = getContainer();
		if ( c == null )
			return;
		c.setCurrentDocumentLocation( location );
	}

	/**
	 * @return The document content as a simple text. <code>null</code> for no current document
	 */
	public String getTextContent() {
		XMLContainer c = getContainer();
		if ( c == null )
			return null;
		return c.getText(); 
	}

	/**
	 * Replace the text content by this one. 
	 * @param content New text content. */
	public void setTextContent( String content ) {
		XMLContainer c = getContainer();
		if ( c != null ) {
			c.setText( content );
		}
	}
	
	/** 
	 * @return The current selection */
	public String getTextSelection() {
		XMLContainer c = getContainer();
		if ( c == null )
			return null;
		return c.getEditor().getSelectedText();
	}

	/**
	 * Replace the selection by this content */
	public void replaceTextSelection( String content ) {
		XMLContainer c = getContainer();
		if ( c != null ) { 
			c.getEditor().replaceSelection( content );
		}
	}

	/** Insert a new content at this location */
	public void insertTextAt( int location, String content ) {
		try {
			XMLContainer c = getContainer();
			if ( c != null ) 
				c.getEditor().getDocument().insertString( location, content, null );
		} catch( Exception exc ) {
		}
	}

	/**
	 * @return The document content as a DOM object, <code>null</code> is returned for no current document 
	 * @throws Exception if the document can't be parsed to DOM
	 * */
	public Document getDOMContent() throws Exception {	
		String content = getTextContent();
		if ( content != null ) {
			DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();		
			return db.parse( new InputSource( new StringReader( content ) ) );
		} else
			return null;
	}

	/** 
	 * @param content The new document content as a DOM object
	 * @throws Exception if the document can't be serialized to a text
	 */
	public void setDomContent( Document content ) throws Exception {
		StringWriter sw = new StringWriter();
		Transformer t = TransformerFactory.newInstance().newTransformer();
		t.transform( new DOMSource( content ), new StreamResult( sw ) );
		setTextContent( sw.toString() );
	}

	/**
	 * @return the Swing editor used by editix, this is useful for particular UI effect */
	public JTextComponent getUIComponent() {
		XMLContainer c = getContainer();
		if ( c == null )
			return null;
		return c.getEditor();
	}

	private XMLContainer getContainer() {	
		return ( XMLContainer )EditixFrame.THIS.getXMLContainer( index );
	}

	/**
	 * Save the current document
	 * @return <code>true</code> if the operation is a success
	 */
	public boolean save() {
		XMLContainer c = getContainer();
		if ( c != null ) {
			SaveAction sa = new SaveAction();
			return sa.save( c );
		}
		return false;
	}

}
