package com.japisoft.framework.dialog.actions;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import javax.swing.AbstractAction;
import javax.swing.Action;

import com.japisoft.framework.SharedProperties;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public abstract class AbstractDialogAction extends AbstractAction implements DialogAction {

	private int id;

	/**
	 * Create a new action with this id
	 * @param id
	 */
	public AbstractDialogAction( int id ) {
		this( id, true );
	}

	/**
	 * @param id a new action id
	 * @param bundleMode decide to reset the label and name with a property file
	 */
	public AbstractDialogAction( int id, boolean bundleMode ) {
		this.id = id;
		if ( bundleMode && SharedProperties.BUNDLE_MODE )
			prepareBundleValues();
		else
			prepareStaticValues();
	}

	/** Called once for setting label, tooltip ... */
	protected void prepareStaticValues() {
	}

	/** Called once for reading properties from bundle */
	protected void prepareBundleValues() {
		try {
			resource = ResourceBundle.getBundle( getBundledName() );
		} catch( MissingResourceException exc ) {
			System.out.println( "Warning : no resource for the action " + getClass().getName() );
		}
		putValue( Action.NAME, getLabel( "NAME", "?" ) );
		putValue( Action.SHORT_DESCRIPTION, getLabel( "SHORT_DESCRIPTION", "?" ) );		
	}

	/** @return the name of the bundle for internationalization */
	protected String getBundledName() {
		return getClass().getName();
	}
	
	private ResourceBundle resource = null;

	/** @return a label from the resource bundle or the default one */	
	protected String getLabel( String name, String def ) {
		if ( resource == null )		
			return def;
		try {
			return resource.getString( name );			
		} catch( MissingResourceException exc ) {
		}
		return def;
	}

	public int getActionId() {
		return id;
	}

	private ActionListener listenerDelegate;
	
	public void setActionDelegate( ActionListener delegate ) {
		this.listenerDelegate = delegate;
	}
	
	public void actionPerformed( ActionEvent e ) {
		if ( listenerDelegate != null )
			listenerDelegate.actionPerformed( e );
	}

	private boolean forFooter = true;

	/** This is a property for not adding the action inside the dialog footer */
	public void setForDialogFooter( boolean forFooter ) {
		this.forFooter = forFooter;
	}

	/** @return <code>true</code> if the action is for the footer. */
	public boolean isForDialogFooter() {
		return forFooter;
	}

	private boolean specialMode = false;
	
	/** @return <code>true</code> if this is a specific action that must appear on the left part */
	public boolean isSpecial() {
		return specialMode;
	}

	/** Update the action status. If <code>true</code>, the action will be on the left part of the footer */
	public void setSpecial( boolean specialMode ) {
		this.specialMode = specialMode;
	}

	private boolean veto = false;
	
	/** Avoid the current dialog to be closed if the action implements the <code>ClosingAction</code> interface */
	public void vetoClosingDialog( boolean veto ) {
		this.veto = veto;
	}

	/** Check if the <code>ClosingAction</code> must be ignored because we don't wish to close
	 * automatically the bound dialog
	 */
	public boolean isVetoClosingDialog() { return veto; }
}
