package com.japisoft.framework.xml.parser.dom;

import com.japisoft.framework.xml.parser.node.*;

import java.util.Enumeration;
import org.w3c.dom.*;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class NamedNodeMapImpl implements NamedNodeMap {
	private FPNode n;

	public NamedNodeMapImpl(FPNode n) {
		super();
		this.n = n;
	}

	/**
	 * Retrieves a node specified by name.
	 * @param name Name of a node to retrieve.
	 * @return A <code>Node</code> (of any type) with the specified name, or 
	 *   <code>null</code> if the specified name did not identify any node in 
	 *   the map. 
	 */
	public Node getNamedItem(String name) {
		// Return an attribute node for the name
		if (n.getAttribute(name) == null)
			return null;
		return new AttrImpl(name, n);
	}

	/**
	 * Adds a node using its <code>nodeName</code> attribute. 
	 * <br>As the <code>nodeName</code> attribute is used to derive the name 
	 * which the node must be stored under, multiple nodes of certain types 
	 * (those that have a "special" string value) cannot be stored as the names 
	 * would clash. This is seen as preferable to allowing nodes to be aliased.
	 * @param arg A node to store in a named node map. The node will later be 
	 *   accessible using the value of the <code>nodeName</code> attribute of 
	 *   the node. If a node with that name is already present in the map, it 
	 *   is replaced by the new one.
	 * @return If the new <code>Node</code> replaces an existing node with the 
	 *   same name  the previously existing <code>Node</code> is returned, 
	 *   otherwise <code>null</code> is returned.
	 * @exception DOMException
	 *   WRONG_DOCUMENT_ERR: Raised if <code>arg</code> was created from a 
	 *   different document than the one that created the 
	 *   <code>NamedNodeMap</code>.
	 *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this 
	 *   <code>NamedNodeMap</code> is readonly.
	 *   <br>INUSE_ATTRIBUTE_ERR: Raised if <code>arg</code> is an 
	 *   <code>Attr</code> that is already an attribute of another 
	 *   <code>Element</code> object. The DOM user must explicitly clone 
	 *   <code>Attr</code> nodes to re-use them in other elements.
	 */
	public Node setNamedItem(Node arg) throws DOMException {
		if (arg instanceof Attr) {
			Attr att = (Attr) arg;
			String name = att.getName();
			if (name == null)
				return null;
			if (n.getAttribute(name) != null) {
				// Reset value
				n.setAttribute(name, att.getValue());
			}
			return arg;
		} else
			throw new DOMExceptionImpl(
				DOMException.NO_MODIFICATION_ALLOWED_ERR,
				"Invalid argument : need an Attr node type ");
	}

	/**
	 * Removes a node specified by name. If the removed node is an 
	 * <code>Attr</code> with a default value it is immediately replaced.
	 * @param name The name of a node to remove.
	 * @return The node removed from the map or <code>null</code> if no node 
	 *   with such a name exists.
	 * @exception DOMException
	 *   NOT_FOUND_ERR: Raised if there is no node named <code>name</code> in 
	 *   the map.
	 */
	public Node removeNamedItem(String name) throws DOMException {
		n.setAttribute(name, null);
		return null;
	}

	/**
	 * Returns the <code>index</code>th item in the map. If <code>index</code> 
	 * is greater than or equal to the number of nodes in the map, this returns 
	 * <code>null</code>.
	 * @param index Index into the map.
	 * @return The node at the <code>index</code>th position in the 
	 *   <code>NamedNodeMap</code>, or <code>null</code> if that is not a valid 
	 *   index.
	 */
	public Node item(int index) {
		return new AttrImpl(n.getViewAttributeAt(index), n);
	}

	/**
	 * The number of nodes in the map. The range of valid child node indices is 
	 * 0 to <code>length-1</code> inclusive. 
	 */
	public int getLength() {
		return n.getViewAttributeCount();
	}

	public Node getNamedItemNS(String namespaceURI, String localName) {
		throw new RuntimeException("Not supported");
	}

	public Node setNamedItemNS(Node arg) throws DOMException {
		throw new RuntimeException("Not supported");
	}

	public Node removeNamedItemNS(String namespaceURI, String localName)
		throws DOMException {
		throw new RuntimeException("Not supported");
	}

}

// NamedNodeMapImpl ends here
