package com.japisoft.xmlform.component;

import java.awt.Dimension;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JMenuItem;

import com.japisoft.framework.ApplicationModel;
import com.japisoft.framework.app.toolkit.Toolkit;
import com.japisoft.framework.xml.grammar.GrammarNode;
import com.japisoft.framework.xml.grammar.GrammarText;
import com.japisoft.framework.xml.grammar.GrammarType;
import com.japisoft.xmlform.component.container.XMLFormContainer;
import com.japisoft.xmlform.component.editable.XMLEnumComponent;
import com.japisoft.xmlform.component.editable.XMLFormTextComponent;
import com.japisoft.xmlform.designer.XmlFormModel;
import com.japisoft.xmlform.designer.data.GrammarNodeTreeNode;
import com.japisoft.xmlform.designer.library.ComponentDescriptor;
import com.japisoft.xmlform.designer.library.ComponentDescriptorModel;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class XMLFormComponentFactory {

	private boolean editableMode = false;
	private ComponentContext context = null;
	
	public XMLFormComponentFactory( 
			boolean editableMode, 
			ComponentContext context ) {
		super();
		this.editableMode = editableMode;
		this.context = context;
	}

	public XMLFormContainer newRootContainer() {
		XMLFormContainer container =
			new XMLFormContainer( editableMode, context );
		container.setTopComponent( true );
		return container;
	}

	public AbstractXMLFormComponent newComponentFromTreeNode( GrammarNodeTreeNode nt ) {
		
		GrammarNode n = nt.getSource();

		GrammarType type = n.getType();
		
		if ( !"complex".equals( type.getType() ) ) {
				
			List<GrammarNode> values = type.getValues();
			if ( values != null ) {
				List<String> r = new ArrayList<String>();
				for ( GrammarNode gn : values ) {
					if ( gn instanceof GrammarText ) {
						r.add( ( ( GrammarText )gn ).getValue() );
					}
				}
				if ( r.size() > 0 ) {
					XMLEnumComponent c = 
						new XMLEnumComponent( true, context );
					String[] ss = new String[ r.size() ];
					for ( int i = 0; i < r.size(); i++ )
						ss[ i ] = r.get( i );
					c.setValues( ss );
					return c;
				}
			}

		}
		
		return newComponentFromTreeNode( nt.toXPath() );
	}

	public AbstractXMLFormComponent newComponentFromTreeNode( String xpath ) {

		AbstractXMLFormComponent component = null;
		
		if ( xpath.endsWith( "text()" ) || xpath.contains( "@" ) ) {
			boolean multiLine = xpath.endsWith( "text()" );
			component = new XMLFormTextComponent( true, context, multiLine );
		} else
			component = new XMLFormContainer( true, context );

		return component;
	}

	public AbstractXMLFormComponent newComponentFromDescriptor( String name ) {
		try {
			ComponentDescriptor cd = 
				ComponentDescriptorModel.getComponentDescriptor( name );
			if ( cd == null )
				return null;
			return cd.create( true, context );
		} catch (Exception e) {
			ApplicationModel.debug( e );
			return null;
		}
	}

	public JButton newAddDeleteComponent() {
		JButton b = 
			new JButton(
					Toolkit.getImageIcon( 
						"images/nav_down_blue.png" ) );
		b.setBorderPainted( false );
		b.setContentAreaFilled( false );
		b.setMargin( null );
		b.setPreferredSize( new Dimension( 10, 10 ) );
		return b;
	}

	public JMenuItem newAddMenuItem() {
		JMenuItem b = 
			new JMenuItem(
					"Add a new item",
					Toolkit.getImageIcon( 
						"images/element_add.png" ) );
		b.setActionCommand( "add" );
		return b;
	}

	public JMenuItem newDeleteMenuItem() {
		JMenuItem b = 
			new JMenuItem(
					"Delete this Item",
					Toolkit.getImageIcon( "images/element_delete.png" ) );
		b.setActionCommand( "delete" );
		return b;		
	}

	public AbstractXMLFormComponent newComponent( 
			String className, 
			HashMap<String,Object> properties, 
			boolean editing, 
			ComponentContext context ) throws Exception {
		
		AbstractXMLFormComponent component = null;
		if ( XMLFormContainer.class.getName().equals( className ) ) {
			component = new XMLFormContainer( editing, context );
		} else
		if ( XMLFormTextComponent.class.getName().equals( className ) ) {
			boolean multiLine = false;
			if ( properties != null ) {
				if ( properties.containsKey( "multiLine" ) )
					multiLine = ( Boolean )properties.get( "multiLine" );
			}
			component = new XMLFormTextComponent( editing, context, multiLine );
		} else
		if ( XMLFormLabelComponent.class.getName().equals( className ) )
			component = new XMLFormLabelComponent( editing, context );
		else
		if ( XMLFormSeparatorComponent.class.getName().equals( className ) ) 
			component = new XMLFormSeparatorComponent( editing, context );
		else
		if ( XMLEnumComponent.class.getName().equals( className ) )
			component = new XMLEnumComponent( editing, context );
		else
			throw new Exception( "Unknown component " + className + "?" );
		
		// Apply properties
		if ( properties != null ) {
			for ( String key : properties.keySet() ) {
				String method = "set" + ( Character.toUpperCase( key.charAt( 0 ) ) ) +
					key.substring( 1 );
				
				Object v = properties.get( key );
				try {
					
					Class cl = v.getClass();
					if ( cl == Boolean.class )
						cl = boolean.class;
					if ( cl == Integer.class )
						cl = int.class;
					
					Method m = 
						component.getClass().getMethod( method, cl );
					m.invoke( component, v );
					
				} catch( Exception e ) {
					XmlFormModel.debug( e );
				}
			}
		}
		return component;
	}

}
